use std::{
    cell::RefCell,
    collections::{HashMap, HashSet},
    rc::Rc,
};

pub struct GlobalNamespace {
    // types
    // constants
    // groups (maybe global data namespace that has groups as subset?)
    types: RefCell<Vec<InternalType>>,
    traits: RefCell<Vec<InternalTrait>>,
    type_namespace: RefCell<HashMap<String, TypeNamespaceId>>,
    commands: RefCell<Vec<InternalCommand>>,
    data_namespace: RefCell<HashMap<String, DataNamespaceId>>,
}

enum TypeNamespaceId {
    Types(usize),
    Traits(usize),
}

enum DataNamespaceId {
    Commands(usize),
    Globals(usize),
}

impl GlobalNamespace {
    pub fn init() -> Self {
        Self {
            types: RefCell::new(Vec::new()),
            traits: RefCell::new(Vec::new()),
            type_namespace: RefCell::new(HashMap::new()),
            commands: RefCell::new(Vec::new()),
            data_namespace: RefCell::new(HashMap::new()),
        }
    }
    pub fn register_type(&self, name: &str) -> Result<Type, TypeRegistrationError> {
        if self.type_namespace.borrow().contains_key(name) {
            Err(TypeRegistrationError::NameAlreadyExists)
        } else {
            self.types.borrow_mut().push(InternalType {
                traits: RefCell::new(HashSet::new()),
                name: name.to_owned(),
            });
            let id = self.types.borrow().len() - 1;
            let _ = self
                .type_namespace
                .borrow_mut()
                .insert(name.to_owned(), TypeNamespaceId::Types(id));
            Ok(Type {
                id,
                namespace: self,
            })
        }
    }

    pub fn register_trait(&self, name: &str) -> Result<Trait, TypeRegistrationError> {
        if self.type_namespace.borrow().contains_key(name) {
            Err(TypeRegistrationError::NameAlreadyExists)
        } else {
            self.traits.borrow_mut().push(InternalTrait {
                types: RefCell::new(HashSet::new()),
                name: name.to_owned(),
            });
            let id = self.traits.borrow().len() - 1;
            let _ = self
                .type_namespace
                .borrow_mut()
                .insert(name.to_owned(), TypeNamespaceId::Traits(id));
            Ok(Trait {
                id,
                namespace: &*self,
            })
        }
    }

    pub fn register_command(
        &self,
        name: &str,
        input: Option<TypeDef>,
        output: Option<TypeDef>,
    ) -> Result<Command, TypeRegistrationError> {
        if self.data_namespace.borrow().contains_key(name) {
            Err(TypeRegistrationError::NameAlreadyExists)
        } else {
            self.commands.borrow_mut().push(InternalCommand {
                name: name.to_owned(),
                input: input.map(|def| def.into()),
                output: output.map(|def| def.into()),
            });
            let id = self.traits.borrow().len() - 1;
            let _ = self
                .data_namespace
                .borrow_mut()
                .insert(name.to_owned(), DataNamespaceId::Commands(id));
            Ok(Command {
                id,
                namespace: self,
            })
        }
    }

    pub fn get_type_by_name(&self, name: &str) -> Option<Type> {
        if let Some(TypeNamespaceId::Types(id)) = self.type_namespace.borrow().get(name) {
            Some(Type {
                id: *id,
                namespace: self,
            })
        } else {
            None
        }
    }

    pub fn get_trait_by_name(&self, name: &str) -> Option<Trait> {
        if let Some(TypeNamespaceId::Traits(id)) = self.type_namespace.borrow().get(name) {
            Some(Trait {
                id: *id,
                namespace: self,
            })
        } else {
            None
        }
    }
}

#[derive(Debug)]
pub enum TypeRegistrationError {
    NameAlreadyExists,
}

pub struct Type<'a> {
    id: usize,
    namespace: &'a GlobalNamespace,
}

impl<'a> Type<'a> {
    pub fn add_trait(&self, to_add: &'a Trait) {
        let internal_self = &self.namespace.types.borrow()[self.id];

        internal_self.traits.borrow_mut().insert(to_add.id);
        self.namespace.traits.borrow_mut()[to_add.id]
            .types
            .borrow_mut()
            .insert(self.id);
    }

    pub fn has_trait(&self, to_check: &'a Trait) -> bool {
        self.namespace.types.borrow()[self.id]
            .traits
            .borrow()
            .contains(&to_check.id)
    }
}

struct InternalType {
    traits: RefCell<HashSet<usize>>,
    name: String,
}

pub struct Trait<'a> {
    id: usize,
    namespace: &'a GlobalNamespace,
}

struct InternalTrait {
    // make resolution easier
    types: RefCell<HashSet<usize>>,
    name: String,
}

struct Command<'a> {
    id: usize,
    namespace: &'a GlobalNamespace,
}

struct InternalCommand {
    name: String,
    // gosh this is hacky
    input: Option<InternalTypeDef>,
    output: Option<InternalTypeDef>,
}

/// the thing used for type annotations etc
/// stores type ids
pub enum TypeDef<'a> {
    /// T
    Type(Type<'a>),
    Trait(Trait<'a>),
    /// [ T ]
    List(Vec<TypeDef<'a>>),
    /// { name: T } or in group declarations [ name: T ]
    Record(Vec<(String, TypeDef<'a>)>),
}

enum InternalTypeDef {
    Single(TypeNamespaceId),
    List(Vec<InternalTypeDef>),
    Record(Vec<(String, InternalTypeDef)>),
}

impl From<TypeDef<'_>> for InternalTypeDef {
    fn from(value: TypeDef) -> Self {
        match value {
            TypeDef::Type(val) => Self::Single(TypeNamespaceId::Types(val.id)),
            TypeDef::Trait(val) => Self::Single(TypeNamespaceId::Traits(val.id)),
            TypeDef::List(list) => Self::List(list.into_iter().map(|def| def.into()).collect()),
            TypeDef::Record(rec) => Self::Record(
                rec.into_iter()
                    .map(|(name, typ)| (name, typ.into()))
                    .collect(),
            ),
        }
    }
}

// what is a type apart from the obvious?
// - it has Traits (integer might have the trait numeric for example)-> also need trait system, should share a namespace tho?

// spec for add
//
// inputs: [ numeric numeric ]
// outputs: numeric
//
